import threading
import time
import logging

class LoadTts(threading.Thread):
    def __init__(self, ext_instance, ml):
        threading.Thread.__init__(self)
        self.ml = ml
        self.ext_instance = ext_instance
        self.hook_instances = self.ml.get_hook_instances()


    def run(self):

        while "Tts" not in self.hook_instances:
            logging.info("Waiting for TTS hook...")
            self.hook_instances = self.ml.get_hook_instances() # Update hook_instances
            time.sleep(1)

        logging.debug("TTS hook found!")
        while len(self.hook_instances['Tts']) == 0:
            logging.info("Waiting for TTS module...")
            self.hook_instances = self.ml.get_hook_instances() # Update hook_instances
            time.sleep(1)

        logging.info("TTS module found! "+type(self.hook_instances["Tts"][0]).__name__)
        if len(self.hook_instances["Tts"]) > 1:
            logging.warning("Multiple TTS modules found. Only the first one will be used.: "+type(self.hook_instances["Tts"][0]).__name__)

        self.ext_instance.tts = self.hook_instances["Tts"][0]


import threading

import src.daisy_llm.ModuleLoader as module_loader
import src.daisy_llm.ContextHandlers as context_handlers

ch = context_handlers('daisy.db')

#Instantiate ModuleLoader and ContextHandlers for global use by front-ends
ml = module_loader(ch, "modules", "configs.yaml")
update_modules_loop_thread = threading.Thread(target=ml.update_modules_loop)
update_modules_loop_thread.start()

ml.process_main_start_instances()


# Import the necessary modules and classes that define the package
from .SoundManager import SoundManager
from .ChatSpeechProcessor import ChatSpeechProcessor
from .ModuleLoader import ModuleLoader
from .Chat import Chat
from .ContextHandlers import ContextHandlers
from .ConnectionStatus import ConnectionStatus
from .LoadTts import LoadTts

# Define the package metadata
__name__ = 'daisy_llm'
__version__ = '0.0.4'
__author__ = 'Myra Krusemark'
__email__ = 'daisy_llm_tools@myrakrusemark.com'

# Export the classes and metadata for use by other modules and packages
__all__ = ['Chat', 'ContextHandlers', 'ModuleLoader', 'ChatSpeechProcessor', 'SoundManager', 'ConnectionStatus', 'LoadTts', '__name__', '__version__', '__author__', '__email__']



from typing import Dict, List, Optional
import sys

from ruamel.yaml import YAML
yaml = YAML()

# Define global variables
yaml_print_text = True
with open("configs.yaml", "r") as f:
    configs = yaml.load(f)
if "print_text" in configs:
    yaml_print_text = configs["print_text"]

TEXT_COLOR_MAPPING = {
    "blue": "36;1",
    "yellow": "33;1",
    "pink": "38;5;200",
    "green": "32;1",
    "red": "31;1",
}

TEXT_STYLE_MAPPING = {
    "bold": "1",
    "italic": "3",
}


def get_color_mapping(
    items: List[str], excluded_colors: Optional[List] = None
) -> Dict[str, str]:
    """Get mapping for items to a support color."""
    colors = list(TEXT_COLOR_MAPPING.keys())
    if excluded_colors is not None:
        colors = [c for c in colors if c not in excluded_colors]
    color_mapping = {item: colors[i % len(colors)] for i, item in enumerate(items)}
    return color_mapping

def get_colored_text(text: str, color:[str] = None, style: Optional[str] = None) -> str:
    """Get colored text."""
    color_str = TEXT_COLOR_MAPPING[color]
    if style is not None:
        style_str = TEXT_STYLE_MAPPING[style]
        return f"\u001b[{color_str}m\033[{style_str}m{text}\u001b[0m"
    else:
        return f"\u001b[{color_str}m{text}\u001b[0m"

def print_text(text: str, color: Optional[str] = None, end: str = "", style: Optional[str] = None) -> None:
    """Print text with highlighting and optional styling."""
    if yaml_print_text:
        if color is None:
            text_to_print = text
        else:
            text_to_print = get_colored_text(text, color, style)
        print(text_to_print, end=end)


def delete_last_lines(n=1): 
    for _ in range(n): 
        sys.stdout.write('\r\x1b[2K') #Erase the line and move the cursor to the start



import openai
import logging
import nltk.data
import threading
import time
import yaml
import json
import requests
import re

from .ChatSpeechProcessor import ChatSpeechProcessor
from .SoundManager import SoundManager
from .Text import print_text, delete_last_lines
import pprint




class Chat:
	description = "Implements a chatbot using OpenAI's GPT-3 language model and allows for interaction with the user through speech or text."

	def __init__(self, ml=None, ch=None):
		self.ml = ml
		self.ch = ch
		self.csp = ChatSpeechProcessor()
		self.sounds = SoundManager()

		with open("configs.yaml", "r") as f:
			self.configs = yaml.safe_load(f)
		openai.api_key = self.configs["keys"]["openai"]

		nltk.data.load('tokenizers/punkt/english.pickle')

	def request(self,
	     messages,
		 stop_event=None,
		 sound_stop_event=None,
		 tts=None,
		 tool_check=True,
		 model="gpt-3.5-turbo",
		 silent=False,
		 response_label=True
		 ):
		#Handle LLM request. Optionally convert to sentences and queue for tts, if needed.

		#Queues for handling chunks, sentences, and tts sounds
		sentences = [[]]  # create a queue to hold the sentences



		if not stop_event:
			stop_event = threading.Event()
		if not sound_stop_event:
			sound_stop_event = threading.Event()

		#Flags for handling chunks, sentences, and tts sounds
		sentence_queue_canceled = [False]  # use a list to make response_canceled mutable
		sentence_queue_complete = [False]	# use a list to make response_complete mutable

		threads = []  # keep track of all threads created
		text_stream = [""]
		return_text = [""]

		#If tool check enabled, find the appropriate tool, and append the output to the messages list.
		if tool_check:
			response = self.toolform_checker(
				messages=messages, 
				stop_event=stop_event,
				response_label=response_label,
				model='gpt-3.5-turbo'
				)
			if response:

				print_text("SYSTEM:", "red", "", "bold")
				print_text(response, None, "\n")

				#Append to CH_context, and the local messages list.
				self.ch.add_message_object("system", response)
				messages.append(self.ch.single_message_context("system", response, False))

		#pprint.pprint(messages)
		try:
			logging.info("Sending request to OpenAI model...")
			response = openai.ChatCompletion.create(
				model=model,
				messages=messages,
				temperature=0.7,
				stream=True,
				request_timeout=5,
			)

			#Handle chunks. Optionally convert to sentences for sentence_queue, if needed.
			arguments = {
				'response': response,
				'text_stream': text_stream,
				'sentences': sentences,
				'sentence_queue_canceled': sentence_queue_canceled,
				'sentence_queue_complete': sentence_queue_complete,
				'return_text': return_text,
				'stop_event': stop_event,
				'sound_stop_event': sound_stop_event,
				'silent': silent,
				'model': model,
				'response_label': response_label
			}
			t = threading.Thread(target=self.stream_queue_sentences, args=(arguments,))
			t.start()
			threads.append(t)

			if tts:
				self.csp.queue_and_tts_sentences(
					tts=tts, 
					sentences=sentences, 
					sentence_queue_canceled=sentence_queue_canceled, 
					sentence_queue_complete=sentence_queue_complete, 
					stop_event=stop_event, 
					sound_stop_event=sound_stop_event
					)

			while not return_text[0] and not stop_event.is_set():
				time.sleep(0.1)  # wait a bit before checking again

			# return response_complete and return_text[0] when return_text is set

			t.join()

			return return_text[0]
		
		# Handle different types of errors that may occur when sending request to OpenAI model
		except openai.error.Timeout as e:
			logging.error(f"Timeout: {e}")
			#self.csp.tts("TimeoutError Error. Check your internet connection.")
			return False  
		except openai.error.APIConnectionError as e:
			logging.error(f"APIConnectionError: {e}")
			#self.csp.tts("APIConnectionError. Sorry, I can't talk right now.")
			return False  
		except openai.error.InvalidRequestError as e:
			logging.error(f"Invalid Request Error: {e}")
			#self.csp.tts("Invalid Request Error. Sorry, I can't talk right now.")
			return False        
		except openai.APIError as e:
			logging.error(f"API Error: {e}")
			#self.csp.tts("API Error. Sorry, I can't talk right now.")
			return False
		except openai.error.RateLimitError as e:
			logging.error(f"RateLimitError: {e}")
			#self.csp.tts("Rate Limit Error. Sorry, I can't talk right now.")
			return False
		except ValueError as e:
			logging.error(f"Value Error: {e}")
			#self.csp.tts("Value Error. Sorry, I can't talk right now.")
			return False    
		except TypeError as e:
			logging.error(f"Type Error: {e}")
			#self.csp.tts("Type Error. Sorry, I can't talk right now.")
			return False  

	def toolform_checker(
			self, 
			messages, 
			stop_event=None, 
			response_label=True,
			model='gpt-3.5-turbo'
			):
		logging.info("Checking for tool forms...")

		#HOOK: Chat_request_inner
		#Right now, only one hook can be run at a time. If a hook returns a value, the rest of the hooks are skipped.
		#I may update this soon to allow for inline responses (For example: "5+5 is [Calculator: 5+5]")
		if self.ml:
			hook_instances = self.ml.get_hook_instances()
			logging.debug(hook_instances)

			if "Chat_request_inner" in hook_instances:

				#Create a tool-chooser prompt
				prompt = """1. "Tools" contains a list of available tools for you to use.
2. Choose one, or more, or None of the tools that are most useful given the context of the "Conversation".
4. Format your response using JSON, like this: [{"name":"tool_form_name", "arg":"tool_form_argument"}]
5. If you choose more than one tool, create an list. Like this: [{"name":"tool_form_name", "arg":"tool_form_argument"}, {"name":"tool_form_name", "arg":"tool_form_argument"}]
6. If you choose no tools, respond with ["None"].
7. Your response will be parsed in a computer program. Do not include any additional text in your response.
8. "Conversation" starts with the earliest message and ends with the most recent message.
9. If the latest message changes the subject of the conversation, even if an earlier message is still relevant, you may respond with ["None"].

Tools:
"""
				#Add all available tools to the prompt
				for module in self.ml.get_available_modules():
					if "tool_form_name" in module:
						prompt += '{"name":"'
						if "tool_form_name" in module:
							prompt += module["tool_form_name"]+'", "arg":"'
						if "tool_form_argument" in module:
							prompt += module["tool_form_argument"]+'"}\n'
						if "tool_form_description" in module:
							prompt += module["tool_form_description"]+"\n\n"

				#Get the last three messages and add them to the prompt
				prompt += "Conversation:\n"
				last_three_messages = messages[-3:]
				for message in last_three_messages:
					prompt += str(message)+"\n"
				logging.info(prompt)
				message = [{'role': 'system', 'content': prompt}]
				logging.info(prompt)

			

				response = self.request(
					messages=message, 
					stop_event=stop_event, 
					tool_check=False,
					model=model,
					silent=True,
					response_label=response_label
					)

				logging.info("Tool form response: "+str(response))

				#Parse JSON response
				data = None
				start_index = response.find('[')
				if start_index >= 0:
					end_index = response.find(']', start_index) + 1
					json_data = response[start_index:end_index]
					try:
						# Check if the input string matches the expected JSON format
						if not re.fullmatch(r'\[.*\]', json_data):
							# Input string does not match expected format
							logging.warning('Input is not valid JSON')
						else:
							# Attempt to load the input string as JSON
							try:
								data = json.loads(json_data)
								logging.info('Data:' + str(data))
							except json.decoder.JSONDecodeError as e:
								# Input string contains errors, attempt to fix them
								logging.error('JSONDecodeError:', e)
								
								# Search for keys with missing values
								match = json_data.search(json_data)
								if match:
									# Replace missing values with empty strings
									fixed_str = json_data[:match.end()] + '""' + json_data[match.end()+1:]
									logging.warning('Fixed input:', fixed_str)
									try:
										data = json.loads(fixed_str)
										logging.info('Data:'+ str(data))
									except json.decoder.JSONDecodeError:
										logging.error('Could not fix input')
								else:
									logging.error('Could not fix input')
	    
						#data = json.loads(json_data)


					except json.decoder.JSONDecodeError as e:
						logging.error("JSONDecodeError: "+str(e))
						data = None
					if data and data[0] == "None":
						data = None
				else:
					logging.warning("No JSON data found in string.")

				logging.info("Tool form chosen: "+str(data))
				
				prompt = ""
				if data:
					for d in data:
						for module in self.ml.get_available_modules():
							if "tool_form_name" in module:
								if module["tool_form_name"] == d["name"]:
									print_text("Tool: ", "green")
									print_text(module["tool_form_name"] + " (" + d['arg']+")", None, "\n\n")

									class_name = module["class_name"]
									chat_request_inner_hook_instances = self.ml.get_hook_instances()["Chat_request_inner"]
									for instance in chat_request_inner_hook_instances:
										if instance.__class__.__name__ == class_name.split(".")[-1]:
											logging.info("Found instance: "+instance.__class__.__name__)
											result = instance.main(d['arg'], stop_event)

											prompt += """Below is the response from the tool: """+module["tool_form_name"]+". It is the information the user is requesting.  Don't mention the existence of this information. Use it to continue the conversation. If the information is irrelevant to the conversation, then ignore it.\n"
											prompt += "\n"+result+"\n"
				if prompt:
					return prompt
				else:
					logging.info("No data found.")
					return False
		return False


	def stream_queue_sentences(self, arguments_dict):
		response = arguments_dict['response']
		text_stream = arguments_dict['text_stream']
		sentences = arguments_dict['sentences']
		sentence_queue_canceled = arguments_dict.get('sentence_queue_canceled', [False])
		sentence_queue_complete = arguments_dict.get('sentence_queue_complete', [False])
		return_text = arguments_dict['return_text']
		stop_event = arguments_dict['stop_event']
		sound_stop_event = arguments_dict['sound_stop_event']
		silent = arguments_dict['silent']
		model = arguments_dict['model']
		response_label = arguments_dict['response_label']

		collected_chunks = []
		collected_messages = []

		try:
			if not silent and response_label:
				print_text("Daisy ("+model+"): ", "blue", "", "bold")

			for chunk in response:
				if not sentence_queue_canceled[0]:
					if not stop_event.is_set():
						temp_sentences = []
						collected_chunks.append(chunk)
						chunk_message = chunk['choices'][0]['delta']
						collected_messages.append(chunk_message)
						text_stream[0] = ''.join([m.get('content', '') for m in collected_messages])
						logging.debug(text_stream[0])

						if not silent:
							if 'content' in chunk_message:
								print_text(chunk_message['content'])
						
						#Tokenize the text into sentences
						temp_sentences = self.csp.nltk_sentence_tokenize(text_stream[0])
						sentences[0] = temp_sentences  # put the sentences into the queue
					else:
						sentence_queue_canceled[0] = True
						logging.info("Sentence queue canceled")
						return
			print_text("\n\n")
		except requests.exceptions.ConnectionError as e:
			logging.error("stream_queue_sentences(): Request timeout. Check your internet connection.")
			sentence_queue_canceled[0] = True

		time.sleep(0.01)
		sentence_queue_complete[0] = True
		return_text[0] = text_stream[0]
		sound_stop_event.set()
		logging.info("Sentence queue complete")
		return


	def display_messages(self, chat_handlers):
		"""Displays the messages stored in the messages attribute of ContectHandlers."""
		for message in chat_handlers.get_context():
			# Check if the message role is in the list of roles to display
			print(f"{message['role'].upper()}: {message['content']}\n\n")




import os
import importlib.util
import logging
import yaml
import time
import threading
import concurrent.futures


from ruamel.yaml import YAML
yaml = YAML()
yaml.allow_duplicate_keys = True


from .Text import print_text



class ModuleLoader:
	initialized = False
	
	def __init__(self, ch, 
		  configs_yaml="configs.yaml", 
		  modules=[]):
		self.ch = ch
		self.configs_yaml = configs_yaml
		self.passed_modules = modules
	
		if not ModuleLoader.initialized:
			ModuleLoader.initialized = True

			self.start_prompts = []
			self.hook_instances = {}
			self.loaded = False

			self.stop_event = threading.Event()
			self.thread = threading.Thread(target=self.update_modules_loop)

			# Load modules
			self.available_modules = []


			# Load enabled modules from config file
			with open(self.configs_yaml, 'r') as f:
				self.configs = yaml.load(f)
			

	def close(self):
		self.stop_event.set()
		self.thread.join()

	def start(self):
		self.thread.start()
		
	def get_hook_instances(self):
		return self.hook_instances
		
	def get_available_modules(self):
		if not self.loaded:
			self.loaded = True
			logging.info("Updating modules...")

			self.enabled_modules = list(self.passed_modules) #Create a copy so that it does not get updated when appended to.

			try:
				with open(self.configs_yaml, 'r') as f:
					self.configs = yaml.load(f)
			except Exception as e:
				logging.warning(f"Failed to load configs.yaml: {str(e)}")

			if "enabled_modules" in self.configs:
				if self.configs["enabled_modules"]:

					#Add modules in configs.yaml to enabled_modules
					for enabled_module in self.configs["enabled_modules"]:
						if enabled_module not in self.enabled_modules: #Eliminates duplicates
							self.enabled_modules.append(enabled_module)

			#Set all modules as disabled and enable them as they are found
			for available_module in self.available_modules:
				available_module["enabled"] = False

			for module_name in self.enabled_modules:
				module_in_available = False

				# If module is in configs.yaml and available_modules, it is enabled.
				for available_module in self.available_modules:
					if module_name == available_module["class_name"]:
						available_module["enabled"] = True
						module_in_available = True


				# Add the module if it's NOT in available modules
				if not module_in_available:
					# Attempt to import the module, and handle exceptions.
					try:
						module = importlib.import_module(module_name, package=None)
					except ModuleNotFoundError as e:
						logging.warning(f"Failed to load {module_name} due to missing dependency: {str(e)}")
						continue  # Skip this module and proceed with the next one

					# Find all classes in the module, and extract their methods and initialization parameters.
					for name in dir(module):

						if name == module.__name__.split(".")[-1]:
							obj = getattr(module, name)
							if isinstance(obj, type):
								module_hook = getattr(obj, "module_hook", "")

								if module_hook and module_name:
									class_description = getattr(obj, "description", "No description.")
									tool_form_name = getattr(obj, "tool_form_name", None)
									tool_form_description = getattr(obj, "tool_form_description", None)
									tool_form_argument = getattr(obj, "tool_form_argument", None)

									# Add module_hook and enabled attributes to module dictionary
									module_dict = {}
									module_dict["class_name"] = module_name
									module_dict["description"] = class_description
									module_dict["module_hook"] = module_hook
									module_dict["enabled"] = True
									if tool_form_name:
										module_dict["tool_form_name"] = tool_form_name
									if tool_form_description:
										module_dict["tool_form_description"] = tool_form_description
									if tool_form_argument:
										module_dict["tool_form_argument"] = tool_form_argument

									self.available_modules.append(module_dict)


			# Notify status of modules
			for available_module in self.available_modules:
				if available_module['enabled']:
					print_text("MODULE LOADED: ", "green", "", "italic")
					print_text(available_module['class_name']+" to "+available_module['module_hook'], None, "\n")

				else:
					print_text("MODULE REMOVED: ", "red", "", "italic")
					print_text(available_module['class_name']+" from "+available_module['module_hook'], None, "\n")

			#If config.yaml item is no longer available, set enabled to False
			for available_module in self.available_modules:
				if available_module["class_name"] not in self.enabled_modules:
					available_module["enabled"] = False
			self.build_hook_instances()
		return self.available_modules

		
	def build_hook_instances(self):
		
		# Create a new dictionary to keep track of updated hook instances
		updated_hook_instances = {}

		# Iterate over the modules in the order they appear
		for module_name in self.enabled_modules:
			for module in self.available_modules:
				
				if module['class_name'] == module_name:
					if module['module_hook'] not in updated_hook_instances:
						updated_hook_instances[module['module_hook']] = []

					# Check if the instance already exists in hook_instances and use it if found
					existing_instance = None

					#Get the module object
					module_class = importlib.import_module(module_name)
					obj = getattr(module_class, module_name.split(".")[-1])

					#Check if the module already exists
					if module['module_hook'] in self.hook_instances:
						for instance in self.hook_instances[module['module_hook']]:
							if isinstance(obj, type) and instance.__class__.__name__ == module_name.split(".")[-1]:
								existing_instance = instance
								break

					#If so, use it instead
					if existing_instance:
						instance = existing_instance
					else:
						for name in dir(module_class):
							if name == module_class.__name__.split(".")[-1]:
								if isinstance(obj, type):
									instance = obj(self)
									instance.ch = self.ch  # Add self.ch to the instance
									instance.ml = self  # Add self to the instance
								if hasattr(instance, "start") and callable(getattr(instance, "start")):
									instance.start()

					# Add the updated instance to the updated_hook_instances
					updated_hook_instances[module['module_hook']].append(instance)
					

					break



		# Close removed instances from hook_instances
		for hook in self.hook_instances:
			for instance in self.hook_instances[hook]:
				if hook not in updated_hook_instances or instance not in updated_hook_instances[hook]:


					if hasattr(instance, 'close') and callable(getattr(instance, 'close')):
						instance.close()

		#Replace existing object with the new one
		self.hook_instances = updated_hook_instances

	def update_modules_loop(self):
		last_modified_time = 0
		while True:
			current_modified_time = os.path.getmtime("configs.yaml")
			if current_modified_time > last_modified_time:
				self.loaded = False
				self.get_available_modules()

				last_modified_time = current_modified_time

			time.sleep(1)

	def start_update_modules_loop_thread(self):
		self.update_modules_loop_thread = threading.Thread(target=self.update_modules_loop)
		self.update_modules_loop_thread.start()

	def stop_update_modules_loop_thread(self):
		self.update_modules_loop_thread.stop()

	def enable_module(self, module_name):
		logging.info("Enabling module: " + module_name)
		with open(self.configs_yaml, 'r') as f:
			config = yaml.load(f)

		if module_name not in config['enabled_modules']:
			config['enabled_modules'].append(module_name)
			with open(self.configs_yaml, 'w') as f:
				yaml.dump(config, f)

			self.loaded = False
		else:
			logging.warning(module_name + " is already enabled.")
		time.sleep(0.5)
		return self.get_available_modules()

	def disable_module(self, module_name):
		logging.info("Disabling module: " + module_name)
		with open(self.configs_yaml, 'r') as f:
			config = yaml.load(f)

		if module_name in config['enabled_modules']:
			config['enabled_modules'].remove(module_name)
			with open(self.configs_yaml, 'w') as f:
				yaml.dump(config, f)

			self.loaded = False
		else:
			logging.warning(module_name + " is already disabled.")
		time.sleep(0.5)
		return self.get_available_modules()

	def process_main_start_instances(self):

		# Define a function that starts a new thread for a given hook instance
		def start_instance(instance):
			logging.info("Main_start: Running %s module: %s "+instance.__class__.__name__+" "+type(instance).__name__)
			future = executor.submit(instance.main, self, self.ch)
			return future

		# Define a dictionary to keep track of running threads
		running_threads = {}
		stop_event = threading.Event()
		# Create the ThreadPoolExecutor outside the while loop
		with concurrent.futures.ThreadPoolExecutor() as executor:
			# Main loop that watches for changes to hook_instances["Main_start"]
			while True:
				logging.debug("Main_start: Checking for changes...")
				if list(running_threads.keys()):
					future_object = list(running_threads.values())[0]  # get the Future object from the dictionary
					if future_object.exception() is not None:  # check if the Future object has a raised exception
						runtime_error = future_object.exception()  # get the raised exception from the Future object
						logging.error("An error occurred: %s "+str(future_object.exception()))


				hook_instances = self.get_hook_instances()
				# Check if any new hook instances have been added or removed
				if "Main_start" in hook_instances:
					for instance in hook_instances["Main_start"]:
						for module in self.get_available_modules():
							if module['class_name'] == instance.__module__ and instance not in running_threads:
								if module['enabled']:
									future = executor.submit(start_instance, instance)
									running_threads[instance] = future
								else:
									future = running_threads[instance]
									future.cancel()
									del running_threads[instance]


				# Wait for some time before checking for updates again
				time.sleep(1)



import pyaudio
import websockets
import asyncio
import base64
import json
import threading
import time
import re
import string
import pyttsx3
import requests
import logging
import yaml
import nltk.data
import queue
import threading
import time
import queue
import requests
from concurrent.futures import ThreadPoolExecutor

from .SoundManager import SoundManager
import modules.RgbLed as led
from .Text import print_text, delete_last_lines



class ChatSpeechProcessor:
	description = "A class that handles speech recognition and text-to-speech processing for a chatbot."

	def __init__(self):
		# Set up AssemblyAI API key and websocket endpoint
		self.uri = "wss://api.assemblyai.com/v2/realtime/ws?sample_rate=16000"


		# Define global variables
		with open("configs.yaml", "r") as f:
			configs = yaml.safe_load(f)
		self.assembly_ai_api_key = configs["keys"]["assembly_ai"]
		self.tts_speed = 1.0

		self.result_str = ""
		self.new_result_str = ""
		self.result_received = False
		self.sounds = SoundManager()
		self.engine = pyttsx3.init()
		self.led = led.RgbLed()
		self.tokenizer = nltk.data.load('tokenizers/punkt/english.pickle')
		self.elapsed_time = 0
		self.timeout_seconds = 0

		self.tts_queue = queue.Queue()  # create a queue to hold the tts_sounds
		self.tts_queue_complete = [False]	# use a list to make response_complete mutable

		self.threads = []  # keep track of all threads created


	def tts(self, text, tts, stop_event=None, sound_stop_event=None):
		with open("configs.yaml", "r") as f:
			configs = yaml.safe_load(f)
			self.tts_speed = configs["TTS"]["speed"]

		audio = tts.create_tts_audio(text)
		self.sounds.play_sound(audio, 1.0, stop_event, sound_stop_event, self.tts_speed)

	def queue_and_tts_sentences(
			self, 
			tts, 
			sentences, 
			sentence_queue_canceled, 
			sentence_queue_complete, 
			stop_event, 
			sound_stop_event=None
			):

		with ThreadPoolExecutor(max_workers=2) as executor:
			arguments = {
				'tts':tts, 
				'sentences':sentences, 
				'sentence_queue_complete':sentence_queue_complete, 
				'sentence_queue_canceled':sentence_queue_canceled, 
				'tts_queue_complete':self.tts_queue_complete,
				'tts_queue':self.tts_queue,
				'stop_event':stop_event, 
			}
			executor.submit(self.queue_tts_from_sentences, arguments)

			arguments = {
				'play_tts_queue':self.play_tts_queue, 
				'tts_queue':self.tts_queue, 
				'sentence_queue_complete':sentence_queue_complete, 
				'sentence_queue_canceled':sentence_queue_canceled, 
				'tts_queue_complete':self.tts_queue_complete,
				'stop_event':stop_event,
				'sound_stop_event':sound_stop_event, 
			}
			executor.submit(self.play_tts_queue, arguments)
		return
		  
	def queue_tts_from_sentences(self, arguments_dict):
		tts = arguments_dict['tts']
		sentences = arguments_dict['sentences']
		sentence_queue_canceled = arguments_dict.get('sentence_queue_canceled', [False])
		sentence_queue_complete = arguments_dict.get('sentence_queue_complete', [False])
		tts_queue_complete = arguments_dict['tts_queue_complete']
		tts_queue = arguments_dict['tts_queue']
		stop_event = arguments_dict['stop_event']

		tts_queue_complete[0] = False
		sentences_length = 1

		def queue_tts_items(index):
			queued_sentence = temp_sentences[index]
			logging.info("Queued sentence: " + queued_sentence)

			try:
				tts_queue.put(tts.create_tts_audio(queued_sentence))
			except requests.exceptions.HTTPError as e:
				self.tts("HTTP Error. Error creating TTS audio. Please check your TTS account.")
				logging.error(f"HTTP Error: {e}")
			except requests.exceptions.ConnectionError as e:
				self.tts("Connection Error. Error creating TTS audio. Please check your TTS account.")
				logging.error(f"Connection Error: {e}")


		while not stop_event.is_set():
			temp_sentences = sentences[0]
			index = 0

			if not sentences[0]:
				continue

			if len(temp_sentences) > sentences_length:
				logging.debug("Sentences: " + str(sentences))
				sentence_length_difference = len(temp_sentences) - sentences_length

				sentences_length = len(temp_sentences)
				


				for i in range(sentence_length_difference):
					if not sentence_queue_canceled[0]:
						index = (sentence_length_difference-i+1) * -1
						queue_tts_items(index)

			elif sentence_queue_complete[0]:

				#Play a single sentence response
				if len(sentences[0]) == 1:
					logging.info("Single sentence response")
					queued_sentence = sentences[0][0]
					logging.info("Queued sentence: "+queued_sentence)

					try:
						if not sentence_queue_canceled[0]:
							tts_queue.put(tts.create_tts_audio(queued_sentence))
					except requests.exceptions.HTTPError as e:
						self.tts("HTTP Error. Error creating TTS audio. Please check your TTS account.")
						logging.error(f"HTTP Error: {e}")

					tts_queue_complete[0] = True
					logging.info("TTS queue complete: single sentence response")
					return

				#Play the very last sentence
				elif len(sentences[0]) == len(temp_sentences):
					if len(sentences[0][-1]) == len(temp_sentences[-1]):
							logging.debug("last sentence...")

							queue_tts_items(-1)

							tts_queue_complete[0] = True
							logging.info("TTS queue complete")
							return

				#All tts items used
				if tts_queue.empty():
					tts_queue_complete[0] = True
					logging.info("TTS queue complete")
					return

			if sentence_queue_canceled[0] or stop_event.is_set():
				tts_queue_complete[0] = True
				while not tts_queue.empty(): #Empty out the TTS queue so no sounds linger
					tts_queue.get()
				logging.info("TTS queue canceled")
				return
			time.sleep(0.5) #Wait juuuust a bit to prevent sentence overlap
				



	def play_tts_queue(self, arguments_dict):
		play_tts_queue = arguments_dict['play_tts_queue']
		tts_queue = arguments_dict['tts_queue']
		sentence_queue_canceled = arguments_dict.get('sentence_queue_canceled', [False])
		sentence_queue_complete = arguments_dict.get('sentence_queue_complete', [False])
		tts_queue_complete = arguments_dict['tts_queue_complete']
		stop_event = arguments_dict['stop_event']
		sound_stop_event = arguments_dict['sound_stop_event']
		
		tts = []

		# Wait for tts to be generated
		while tts_queue.empty() and not sentence_queue_canceled[0]:
			time.sleep(0.01)

		# Play tts
		while not stop_event.is_set():
			tts = None


			if tts_queue.qsize():
				tts = tts_queue.get(block=True, timeout=0.01)  # get tts from the queue

				#Stop voice assistant "waiting" sound
				if sound_stop_event:
					sound_stop_event.set()
			
				if tts:
					
					# Define global variables
					with open("configs.yaml", "r") as f:
						configs = yaml.safe_load(f)
						if "TTS" in configs:
							if "speed" in configs["TTS"]:
								self.tts_speed = configs["TTS"]["speed"]
					self.sounds.play_sound(tts, 1.0, stop_event, None, self.tts_speed)
			elif tts_queue_complete[0]:
				logging.info("TTS play queue complete")
				return


	async def stt_send_receive(self, stop_event, timeout_seconds=0, sound = None):
		"""Sends audio data to AssemblyAI STT API and receives text transcription in real time using websockets."""

		self.result_str = ""
		self.new_result_str = ""
		self.result_received = False
		self.timeout_seconds = timeout_seconds

		# Set up PyAudio
		FRAMES_PER_BUFFER = 3200
		FORMAT = pyaudio.paInt16
		CHANNELS = 1
		RATE = 16000
		p = pyaudio.PyAudio()
		stream = p.open(
			format=FORMAT,
			channels=CHANNELS,
			rate=RATE,
			input=True,
			frames_per_buffer=FRAMES_PER_BUFFER
		)

		try:
			async with websockets.connect(
				self.uri,
				extra_headers=(("Authorization", self.assembly_ai_api_key),),
				ping_interval=5,
				ping_timeout=20
			) as _ws:
				await asyncio.sleep(0.1)
				logging.info("Receiving SessionBegins ...")
				session_begins = await _ws.recv()
				logging.info(session_begins)
				logging.info("AAI Listening ...")

				async def timeout() -> None:
					start_time: float = time.time()
					self.elapsed_time: float = 0

					while not self.result_received and not stop_event.is_set():
						self.elapsed_time: float = time.time() - start_time
						if stop_event.is_set():
							logging.info("Timeout()")
							break
						if timeout_seconds > 0: # If timeout is 0s, then dont timeout
							if self.elapsed_time > timeout_seconds:
								logging.info("Timeout reached")
								stop_event.set()
								return
						await asyncio.sleep(0.01)

					logging.info("timeout(): Timeout cancelled or result received")

					return


				async def send():
					logging.info("STT Send start")

					while not self.result_received and not stop_event.is_set():
						if stop_event.is_set():
							logging.info("Send(): Cancelled")
							break

						try:
							data = stream.read(FRAMES_PER_BUFFER)
							data = base64.b64encode(data).decode("utf-8")
							json_data = json.dumps({
								"audio_data":str(data), 
								"punctuate": False, 
								"format_text": False
								})
							await _ws.send(json_data)
						except websockets.exceptions.ConnectionClosedError as e:
							logging.error(f"Connection closed with error code {e.code}: {e.reason}")
							break
						except Exception as e:
							logging.exception(f"Unexpected error: {e}")
							break
						await asyncio.sleep(0.01)

					logging.info("send(): STT Send done")
					return
							
				
				async def receive():
					logging.info("STT Receive start")

					while not self.result_received and not stop_event.is_set():
						if stop_event.is_set():
							logging.info("receive(): Cancelled")
							self.result_str = False
							self.result_received = True
							break

						try:
							self.new_result = await asyncio.wait_for(_ws.recv(), timeout=3) #Timeout if connection is lost
							self.result_str_obj = json.loads(self.new_result)

							delete_last_lines()
							print_text("You ("+str(round(self.timeout_seconds - self.elapsed_time))+"s): ", "green", "", "bold")
							print_text(str(self.result_str_obj['text']))

							self.led.turn_on_color_random_brightness(0, 0, 100)  # Random brightness Blue

							# If the message type is FinalTranscript, then we are done
							if self.result_str_obj['message_type'] == "FinalTranscript" and self.result_str_obj['text'] != "":
								#DONE
								logging.info("receive(): STT Receive done")
								logging.info("receive(): You said: "+str(self.result_str_obj['text']))

								self.result_str = self.result_str_obj['text']
								self.result_received = True
								print_text("", None, "\n")

						except asyncio.TimeoutError:
							logging.warning("receive(): Receive timed out")
							self.result_str = False
							self.result_received = True
						except websockets.exceptions.ConnectionClosedError as e:
							logging.error(f"Connection closed with error code {e.code}: {e.reason}")
							self.result_str = False
							self.result_received = True
						except Exception as e:
							logging.exception(f"Unexpected error: {e}")
							self.result_str = False
							self.result_received = True
					logging.info("receive(): STT Receive done")
					return



				#Play start sound
				if sound:
					self.sounds.play_sound_with_thread(sound)

				send_result, receive_result, timeout_result = await asyncio.gather(
					asyncio.shield(timeout()), asyncio.shield(send()), asyncio.shield(receive())
				)
		except websockets.exceptions.ConnectionClosedError as e:
			logging.error(f"Connection closed with error code {e.code}: {e.reason}")
			self.result_str = False
			self.result_received = True


	def stt(self, stop_event, timeout_seconds=0, sound = None):
		"""Calls stt_send_receive in a new thread and returns the final transcription."""
		# Create an event object to signal the thread to stop
		stt_stop_event = threading.Event()

		def watch_results():
			#global result_received
			#global result_str
			while not stt_stop_event.is_set() and not stop_event.is_set():
				if self.result_received:
					logging.info("Result received: %s", self.result_str)
					self.result_received = False

					# Set the event to signal the thread to stop
					stt_stop_event.set()

				time.sleep(0.1)

			# Join the thread to wait for it to finish
			logging.debug("Joining STT thread...")
			thread.join()
			logging.debug("STT Thread stopped")

			return self.result_str


		# Set up AssemblyAI stt_send_receive loop
		#This is a thread in a thread. I think it can be reduced.
		def start_stt_send_receive():
			asyncio.run(self.stt_send_receive(stop_event, timeout_seconds, sound))

		# Create and start the stt_send_receive thread
		thread = threading.Thread(target=start_stt_send_receive)
		thread.start()

		# Start watching results in the main thread
		result_str = watch_results()

		return result_str

	def remove_non_alphanumeric(self, text):
		"""Removes all characters that are not alphanumeric or punctuation."""

		# Create a set of all valid characters
		valid_chars = set(string.ascii_letters + string.digits + "!()',./?+=-_#$%&*@" + ' ')

		# Use a generator expression to filter out any invalid characters
		filtered_text = ''.join(filter(lambda x: x in valid_chars, text))

		# Log the input and output text at the DEBUG level
		logging.debug(f'Removing non-alphanumeric characters from text: {text}')
		logging.debug(f'Filtered text: {filtered_text}')

		return filtered_text

	def remove_non_alpha(self, text):
		"""Removes all non-alphabetic characters (including punctuation and numbers) from a string and returns the modified string in lowercase."""
		if text:
			# Log a debug message with the input string
			logging.debug(f'Removing non-alpha characters from string: {text}')

			# Use regular expression to replace non-alphanumeric characters with empty string
			text = re.sub(r'[^a-zA-Z]+', '', text)

			# Log a debug message with the modified string
			logging.debug(f'Filtered text: {text}')

			# Return the modified string
			return text.lower()
		else:
			return False
		
	def nltk_sentence_tokenize(self, text, language="english"):
		"""Splits a string into sentences using the NLTK sentence tokenizer."""
		# Log a debug message with the input string
		logging.debug(f'Tokenizing string: {text}')

		# Split the string into sentences
		sentences = nltk.sent_tokenize(text, language)

		# Log a debug message with the modified string
		logging.debug(f'Tokenized sentences: {sentences}')

		# Return the modified string
		return sentences




import os
from io import BytesIO
import logging
import threading
import pydub
from pydub import AudioSegment
from pydub.effects import speedup
import simpleaudio
import numpy as np
import pkg_resources

class SoundManager:
    description = "A class for managing sound files in a directory, playing sounds, and stopping playback."
    
    def __init__(self, directory='sounds'):
        self.directory = pkg_resources.resource_filename(__name__, directory)
        self.sounds = {}
        self.current_sound = None
        self.playback = None

        # Load all sound files in the directory
        for filename in os.listdir(self.directory):
            if filename.endswith('.wav') or filename.endswith('.mp3'):
                # Extract filename without extension
                name = os.path.splitext(filename)[0]

                # Load sound file and store in dictionary
                path = os.path.join(self.directory, filename)
                try:
                    sound = pydub.AudioSegment.from_file(path)
                    self.sounds[name] = sound
                except pydub.exceptions.CouldntDecodeError:
                    logging.warning(f"Failed to load sound file '{filename}'")


    def play_sound(self, name_or_bytes, volume=1.0, stop_event=None, sound_stop_event=None, speed=1.0):
        logging.debug("play_sound")
        if isinstance(name_or_bytes, bytes):
            sound = AudioSegment.from_file(BytesIO(name_or_bytes))
        elif isinstance(name_or_bytes, str):
            sound = self.sounds.get(name_or_bytes)
            if sound is None:
                raise ValueError(f"No sound named {name_or_bytes}")
        else:
            raise TypeError(f"Unsupported argument type {type(name_or_bytes)}")

        self.current_sound = sound

        if stop_event is None:
            stop_event = threading.Event()

        # adjust the playback speed of the sound
        if speed != 1.0:
            sound = speedup(sound, speed)

        # create a new thread to play the sound
        self._play_sound_method(sound, volume, stop_event, sound_stop_event)

        self.current_sound = None
        return True
    

    def _play_sound_method(self, sound, volume=1.0, stop_event=None, sound_stop_event=None):
        logging.debug("_play_sound_method")
        # Adjust volume
        scaled_volume = volume * 32767
        sound_array = sound.get_array_of_samples()
        scaled_array = np.round(scaled_volume * np.array(sound_array) / 32767).astype(np.int16)
        sound = sound._spawn(scaled_array)

        # Convert audio segment to raw audio data
        raw_data = sound.raw_data

        # Set audio parameters
        num_channels = sound.channels
        bytes_per_sample = sound.sample_width
        sample_rate = sound.frame_rate

        # Play audio using simpleaudio
        self.playback = simpleaudio.play_buffer(raw_data, num_channels=num_channels, bytes_per_sample=bytes_per_sample, sample_rate=sample_rate)

        # Loop until either stop_event or sound_stop_event is set
        while True:
            if stop_event:
                if stop_event.is_set():
                    # Stop the playback if the stop_event is set
                    self.playback.stop()
                    return False
            if sound_stop_event:
                if sound_stop_event.is_set():
                    # Stop the playback if the sound_stop_event is set
                    self.playback.stop()
                    return True

            # Check if the playback has finished
            if not self.playback.is_playing():
                return True



    def play_sound_with_thread(self, name_or_bytes, volume=1.0, awake_stop_event=None, sound_stop_event=None):
        logging.debug("play_sound_with_thread")
        thread = threading.Thread(target=self.play_sound, args=(name_or_bytes, volume, awake_stop_event, sound_stop_event))
        thread.start()





import logging
import datetime
import json
import sqlite3
import time
import threading
from sqlite3 import Error
import sqlite3
from .Chat import Chat
import re
from ruamel.yaml import YAML
from .Text import print_text
yaml = YAML()

class ConnectionPool:
	def __init__(self, db_path, max_connections=5):
		self.db_path = db_path
		self.max_connections = max_connections
		self.connections = {}
		self.lock = threading.Lock()



	def get_connection(self):
		thread_id = threading.get_ident()
		self.lock.acquire()
		try:
			if thread_id in self.connections:
				return self.connections[thread_id]
			elif len(self.connections) < self.max_connections:
				conn = sqlite3.connect(self.db_path)
				self.connections[thread_id] = conn
				return conn
			else:
				raise Exception("Connection pool exhausted")
		finally:
			self.lock.release()

	def put_connection(self, conn):
		thread_id = threading.get_ident()
		self.lock.acquire()
		try:
			if thread_id in self.connections:
				self.connections[thread_id] = None
			else:
				conn.close()
		finally:
			self.lock.release()


class ContextHandlers:
	description = "A class for handling and managing messages in the chatGPT context object"

	def __init__(self, db_path):
		self.chat = Chat()

		#Get and set conversation_id from configs.yaml
		self.conversation_id = None
		with open("configs.yaml", "r") as f:
			configs = yaml.load(f)
		if 'conversation_id' in configs:
			self.conversation_id = configs.get("conversation_id")
			logging.info("Using conversation id from configs: " + str(self.conversation_id))


		self.db_path = db_path
		self.messages = []
		self.start_prompts = []
		self.connection_pool = ConnectionPool(db_path)

		self.load_context()

	def update_conversation_name_summary(self):
		# Get the name of the current conversation from the LLM
		time.sleep(1)
		logging.info("Updating conversation name and summary...")
		messages = self.get_context_without_timestamp()
		prompt = """
		Please respond with a name, and summary for this conversation.
		1. The name should be a single word or short phrase, no more than 5 words."
		2. The summary should be a short description of the conversation, no more than 2 SHORT sentences.
		3. The output must follow the following JSON format: {"name": name, "summary": summary}
		"""
		messages.append(self.single_message_context('system', prompt, False))

		print_text("Conversation info ("+str(self.conversation_id)+"): ", "yellow")
		response = self.chat.request(
			messages=messages, 
			silent=False, 
			response_label=False
			)

		# Extract the JSON response from the string
		response_match = re.search(r"{.*}", response)
		if response_match:
			response_json = response_match.group(0)
		else:
			logging.error("Invalid response format while setting conversation name and summary")
			return

		# Convert the JSON response to an object
		try:
			response_obj = json.loads(response_json)
		except Exception as e:
			logging.error("Invalid JSON response while setting conversation name and summary: " + str(e))
			return

		# Update the name and summary of the current conversation in the database
		with self.connection_pool.get_connection() as conn:
			cursor = conn.cursor()
			cursor.execute(
				'''UPDATE conversations SET name = ?, summary = ? WHERE id = ?''',
				(response_obj["name"], response_obj["summary"], self.conversation_id)
			)
			conn.commit()

		logging.info("Name and summary updated: " + response_obj["name"])

		return


	def load_context(self):
		self.messages = []
		self.create_conversations_table_if_not_exists()
		with self.connection_pool.get_connection() as conn:

			cursor = conn.cursor()

			# Get the conversation ID
			if not self.conversation_id:
				cursor.execute('''
				SELECT id FROM conversations ORDER BY id DESC LIMIT 1;
				''')
				row = cursor.fetchone()
				if row:
					self.conversation_id = str(row[0])
					logging.info("No conversation id found in configs.yaml, loading latest conversation: " + str(self.conversation_id))

			# If conversation still is not set, create a new conversation ID
			if not self.conversation_id:
				self.conversation_id = str(int(time.time()))
				logging.info("No conversation found, creating new conversation: " + str(self.conversation_id))

				# Set the new conversation ID in configs.yaml
				with open("configs.yaml", "r") as f:
					configs = yaml.load(f)
				configs['conversation_id'] = self.conversation_id
				with open("configs.yaml", "w") as f:
					yaml.dump(configs, f)

			logging.info("Conversation id: " + str(self.conversation_id))

			# Get the messages from the conversation ID
			cursor.execute('''
				SELECT * FROM messages WHERE id = ?
			''', (self.conversation_id,))
			rows = cursor.fetchall()
			if rows:
				for row in rows:
					message = json.loads(row[1])
					self.messages.append(message)
				print_text("Loaded "+ str(len(rows)) + " messages from conversation id: " + str(self.conversation_id), "yellow", "\n")



	def create_conversations_table_if_not_exists(self):
		with self.connection_pool.get_connection() as conn:
			cursor = conn.cursor()
			cursor.execute('''
				CREATE TABLE IF NOT EXISTS messages (
					id TEXT NOT NULL,
					message TEXT NOT NULL
				);
			''')
			cursor.execute('''
				CREATE TABLE IF NOT EXISTS conversations (
					id TEXT PRIMARY KEY,
					name TEXT NOT NULL,
					summary TEXT NOT NULL
				);
			''')

	def save_context(self):
		logging.info("Saving context: " + str(self.conversation_id))
		with self.connection_pool.get_connection() as conn:
			conn.execute('''
				PRAGMA foreign_keys=OFF;
			''')
			conn.execute('''
				BEGIN TRANSACTION;
			''')

			# Insert conversation information if it doesn't already exist
			rows = conn.execute('''
				SELECT COUNT(*) FROM conversations WHERE id = ?;
			''', (self.conversation_id,)).fetchone()[0]
			if rows == 0:
				conn.execute('''
					INSERT INTO conversations (id, name, summary)
					VALUES (?, ?, ?);
				''', (self.conversation_id, "generic name", "generic summary"))

			# Save messages
			conn.execute('''
				DELETE FROM messages WHERE id = ?;
			''', (self.conversation_id,))
			for message in self.messages:
				json_message = json.dumps(message)
				conn.execute('''
					INSERT INTO messages (id, message) VALUES (?, ?);
				''', (self.conversation_id, json_message,))
			conn.execute('''
				COMMIT;
			''')
			row_count = conn.execute('''
				SELECT COUNT(*) FROM messages WHERE id = ?;
			''', (self.conversation_id,)).fetchone()[0]
			logging.info(f"Inserted {row_count} rows for conversation {self.conversation_id}.")


	def get_context(self):
		context = []
		#Append start prompts to messages
		for start_prompt in self.start_prompts:
			context.append(start_prompt)

		for message in self.messages:
			context.append(message)
		return context


	def get_context_without_timestamp(self):
		messages_without_timestamp = []

		for message in self.get_context():
			message_without_timestamp = message.copy()
			del message_without_timestamp['timestamp']
			messages_without_timestamp.append(message_without_timestamp)
		return messages_without_timestamp
	
	def get_conversation_name_summary(self):
		with self.connection_pool.get_connection() as conn:
			cursor = conn.cursor()
			cursor.execute(
				'''SELECT name, summary FROM conversations'''
			)
			rows = cursor.fetchall()
			if rows:
				return [(name, summary) for name, summary in rows]
			else:
				return None


	def single_message_context(self, role, message, incl_timestamp=True):
		if incl_timestamp:
			now = datetime.datetime.now()
			timestamp = now.strftime("%Y-%m-%d %H:%M:%S")
			return {'role': role, 'timestamp': timestamp, 'content': str(message)}
		else:
			return {'role': role, 'content': str(message)}
		
	def add_start_propmpt(self, role="system", message=""):
		start_prompt = self.single_message_context(role, message)
		self.start_prompts.append(start_prompt)

	def add_message_object(self, role, message):
		logging.debug("Adding " + role + " message to context")
		now = datetime.datetime.now()
		timestamp = now.strftime("%Y-%m-%d %H:%M:%S")
		new_message = {'role': role, 'timestamp': timestamp, 'content': str(message)}
		self.messages.append(new_message)
		self.save_context()
		logging.debug(self.messages)

	def add_message_object_at_start(self, role, message):
		logging.debug("Appending " + role + " message at start of context")
		now = datetime.datetime.now()
		timestamp = now.strftime("%Y-%m-%d %H:%M:%S")
		new_message = {'role': role, 'timestamp': timestamp, 'content': str(message)}
		self.messages.insert(0, new_message)
		self.save_context()
		logging.debug(self.messages)

	def remove_last_message_object(self):
		if self.messages:
			self.messages.pop()
			self.save_context()

	def get_last_message_object(self, user_type=None):
		if user_type:
			for message in reversed(self.messages):
				if message['role'] == user_type:
					return message
		else:
			if self.messages:
				return self.messages[-1]
		return False

	def replace_last_message_object(self, message, user_type=None):
		if user_type:
			for i in reversed(range(len(self.messages))):
				if self.messages[i]['role'] == user_type:
					self.messages[i]['content'] = message
					self.save_context()
					return
		elif message and self.messages:
			self.messages[-1]['content'] = message
			self.save_context()

	def delete_message_at_index(self, index):
		try:
			index = int(index)
			if index < len(self.messages) and index >= 0:
				self.messages.pop(index)
				self.save_context()
				return True
		except ValueError:
			pass
		return False

	def update_message_at_index(self, message, index):
		try:
			index = int(index)
			if index < len(self.messages) and index >= 0:
				self.messages[index]['content'] = message
				now = datetime.datetime.now()
				self.messages[index]['timestamp'] = now.strftime('%Y-%m-%d %H:%M:%S')
				self.save_context()
		except ValueError:
			pass
		return False



import platform
import subprocess
import threading
import logging
import time


class ConnectionStatus(threading.Thread):
    description = "A class that checks the status of internet connectivity."

    def __init__(self):
        self.is_connected = False
        self.last_loss_time = None

    def check_internet(self, stop_event, awake_stop_event):
        """Creates and starts a new CheckInternetThread instance, waits for it to complete or timeout, and returns a boolean value indicating if internet connectivity is available."""

        while not stop_event.is_set():
            try:
                # Ping google.com and check if we receive any response
                logging.debug('Pinging google.com')

                if platform.system() == 'Windows':
                    output = subprocess.check_output(['ping', '-n', '5', '-w', '5000', 'google.com'], stderr=subprocess.PIPE)
                else:
                    output = subprocess.check_output(['ping', '-c', '5', '-W', '5', 'google.com'], stderr=subprocess.PIPE)

                if "Reply from" in output.decode():
                    self.is_connected = True

                    # Log a message to indicate that the Internet is available
                    logging.debug('Internet connection is available')
                    self.last_loss_time = None
                else:
                    self.is_connected = False
                    now = time.monotonic()
                    if self.last_loss_time is None:
                        self.last_loss_time = now
                    elif now - self.last_loss_time >= 10:
                        awake_stop_event.set()
                        logging.error("No Internet connection for at least 10 seconds.")
                    else:
                        awake_stop_event.clear()
                    # Log a warning message to indicate that the Internet is not available
                    logging.debug('Internet connection is not available')

            except (subprocess.CalledProcessError, subprocess.TimeoutExpired):
                self.is_connected = False
                now = time.monotonic()
                if self.last_loss_time is None:
                    self.last_loss_time = now
                elif now - self.last_loss_time >= 10:
                    awake_stop_event.set()
                    logging.error("No Internet connection for at least 10 seconds.")
                else:
                    awake_stop_event.clear()
                # Log a warning message to indicate that the Internet is not available
                logging.debug('Internet connection is not available')
            except Exception as e:
                # general exception handling
                logging.exception(e)

            # Check if the thread is connected and cancel the loop if it is
            if not self.is_connected:
                awake_stop_event.clear()
            time.sleep(1)



from src.daisy_llm.ChatSpeechProcessor import ChatSpeechProcessor
from unittest.mock import MagicMock
from src.daisy_llm.ContextHandlers import ContextHandlers
from src.daisy_llm.SoundManager import SoundManager
from src.daisy_llm.ModuleLoader import ModuleLoader
from src.daisy_llm.Chat import Chat
import os

test_messages=[{"role": "system", "content": "Daisy"}]

csp = ChatSpeechProcessor()
ch = ContextHandlers(test_messages)
#chat = Chat(os.environ["API_KEY"], test_messages)
sm = SoundManager("sounds/")
ml = ModuleLoader("modules")

def test_ModuleLoader(capsys):
    with capsys.disabled():
        print(ml.get_available_classes_json())

def test_SoundManager(capsys):
    with capsys.disabled():
        print("Press ESC to stop the sound")
        sm.play_sound("waiting")
        print("Sound Stopped")
        input_text = input('Did the sound stop when you pressed ESC? (y/n)')
        if input_text == 'y':
            passed = True
        assert passed == True

def test_ChatSpeechProcessor_text():
    assert csp.remove_non_alpha("!abc. 123?") == 'abc'
    assert csp.remove_non_alphanumeric("!abc. /?{]^123?[]{}|^") == "!abc. /?123?"
    assert str(csp.split_text_for_google_tts(csp.remove_non_alphanumeric("In publishing and graphic design, Lorem ipsum is a placeholder text commonly used to demonstrate the visual form of a document or a typeface without relying on meaningful content. Lorem ipsum may be used as a placeholder before final copy is available."))) == "['In publishing and graphic design, Lorem ipsum is a placeholder text commonly used to demonstrate the visual form of a document or a typeface without relying on meaningful content. Lorem ipsum may be', 'used as a placeholder before final copy is available.']"

def test_ChatSpeechProcessor(capsys):
    text = ""
    while True:
        with capsys.disabled():
            print("Say: Daisy")
            passed = csp.listen_for_wake_word()
            if passed:
                break
    assert passed == True

    with capsys.disabled():
        text = "This is a test"
        csp.tts(text)
        input_text = input('Did you hear "This is a test."? (y/n)')
        if input_text == 'y':
            passed = True

        assert passed == True

def test_ContextHandlers(capsys):
    test_output=[{"role": "system", "content": "Daisy"},
                    {"role": "user", "content": "testing..."}]
    ch.add_message_object('user', 'testing...')
    assert test_messages == test_output

    ch.add_message_object('user', 'This is a test. Please reply only with "Testing, 1, 2, 3."')
    with capsys.disabled():
        text = chat.chat()
        assert text == "Testing, 1, 2, 3."

def test_Chat(capsys):
    modules = [{"name":"GoogleScraper", "module_hook":"Chat_chat_inner"}]
    with capsys.disabled():

        chat.import_and_instantiate_classes(modules, "modules")
        print("Instances: "+str(chat.Chat_chat_inner_instances))

    assert str(chat.Chat_chat_inner_instances).find("GoogleScraper")



from daisy_llm import ModuleLoader, ContextHandlers

ch = ContextHandlers('daisy.db')
ml = ModuleLoader(ch,
                  configs_yaml="configs.yaml",
                  modules=["modules.Daisy.Daisy"]
                  )


#Start ModuleLoader dynamic checker
ml.start_update_modules_loop_thread()

#Start front end sub processes
ml.process_main_start_instances()

#When front ends are done, or signal received, stop the loop
ml.stop_update_modules_loop_thread()